# Level class For Kourier
# Ben Delacour Spring 2006

import glob
import helpers
import math
import os, os.path
import pygame
import random
import string
import sys
from pygame.locals import *
from Cargen        import Cargen
from Car           import Car




class Level:
    """A Level class for Kourier"""

    def __init__(self):
        self.unknown_commands_arr= []
        self.part_description= "[level]"
        self.description= "[level]"
        self.level_name= "[level0]"
        self.next_level= "[level1]"
        self.min_car_speed= 22
        self.max_car_speed= 22
        self.tile_filename_arr= []
        self.tile_imgnrect_arr= []
        self.tile_floor1_arr2d=  []     # floor1: indicies into "imgage and rect" array
        self.tile_floor2_arr2d=  []     # floor2: indicies into "imgage and rect" array
        self.tile_floor_s_arr2d= []     # floor3: indicies into "imgage and rect" array
        self.car_floor_arr2d= []        # car collision detection floor array
        self.explosions_arr2d= []       # array of: [x loc, y loc, explosion existance time, explosion poly, poly color]
        self.explosion_size= 6          # should equate roughly to frames displayed
        self.explosion_size_inc= .3
        self.frame_number= 0            # used to invalidate the car collision detection array implicitly
        self.car_floor_fraction= 4      # eg: if 4, the car floor will be 1/4 as wide & tall as other floors

        self.cur_parcel= 0
        self.cur_mailbox= 0
        self.parcel_count= 0
        self.parcel_pos_arr=  [[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1]] # positions of the tiles
        self.mailbox_pos_arr= [[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1]] # positions of the tiles
        self.bonus_time_arr= [] # minimum times for bonus money
        self.last_bonus= 0.00
        self.last_bonus_str= "$0.00"
        self.parcel_delivery_time= 0
        self.parcel_pick_up_time= 0
        self.can_get_bonus_money= True
        self.cargen_arr= []
        self.car_imgnrect_arr= []
        self.car_arr= []

        self.screen_pos= [0, 0]
        self.screen_size= [640, 480]
        self.tiledir= "tiles"
        self.tile_dimen= 64
        self.width= 0
        self.height= 0



    def set_grappled_car(self, index):
        i= index
        self.car_arr= [self.car_arr[i]]+ self.car_arr[0: i]+ self.car_arr[i+ 1:]
        self.car_arr[0].grappled= True


    def release_grappled_car(self):
        self.car_arr[0].grappled= False




    def get_grappled_car(self):
        return self.car_arr[0]



    def load(self, _level_name, initial_updates= 40*20):
        # Loads the specified level
        random.seed()
        self.level_name= _level_name
        self.unknown_commands_arr= []
        self.min_car_speed= 18
        self.max_car_speed= 24
        self.tile_filename_arr= []
        self.tile_imgnrect_arr= []
        self.tile_floor1_arr2d=  []     # floor1: indicies into "imgage and rect" array
        self.tile_floor2_arr2d=  []     # floor2: indicies into "imgage and rect" array
        self.tile_floor_s_arr2d= []     # floor3: indicies into "imgage and rect" array
        self.car_floor_arr2d= []        # car collision detection floor array
        self.frame_number= 0            # used to invalidate the car collision detection array implicitly
        self.car_floor_fraction= 4      # eg: if 4, the car floor will be 1/4 as wide & tall as other floors

        self.cur_parcel= 0
        self.cur_mailbox= 0
        self.parcel_count= 0
        self.parcel_pos_arr=  [[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1]] # positions of the tiles
        self.mailbox_pos_arr= [[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1],[-1,-1]] # positions of the tiles
        self.bonus_time_arr= [] # minimum times for bonus money
        self.last_bonus= 0.00
        self.last_bonus_str= "$0.00"
        self.parcel_delivery_time= 0
        self.parcel_pick_up_time= 0
        self.can_get_bonus_money= True
        self.cargen_arr= []
        self.car_imgnrect_arr= []
        self.car_arr= []

        self.screen_pos= [0, 0]
        self.screen_size= [640, 480]
        self.tiledir= "tiles"
        self.tile_dimen= 64

        self.player_start_pos= [0, 0]
        self.player_checkpoint_pos= [0, 0]

        _unknown_commands_arr= []
        full_name= os.path.join("levels", _level_name)
        if os.path.exists(full_name)== False:
                print "Error in load_level: map %s does not exist." % full_name
                
        #print "Loading level in load_level: %s" % full_name
        fin= open(full_name, 'r')

        whole= fin.read()
        line_arr= string.split(whole, '\n')
        for i in range(0, len(line_arr)):

            colin= line_arr[i].find(":")
            if colin == -1:
                continue

            [description, value]= string.split(line_arr[i], ":")
            if len(description) > 5 and description[0:4]== "tile":
                self.tile_filename_arr.append(os.path.join(self.tiledir, value[1:len(value)]))

  
            elif len(description) > 9 and description== "description":
                self.description= value[1:]

            elif len(description) > 9 and description== "part_description":
                self.part_description= value[1:]
        
            elif len(description) > 9 and description== "next_level":
                self.next_level= value[1:]

            elif len(description) > 9 and description== "min_car_speed":
                self.min_car_speed= int(value[1:])

            elif len(description) > 9 and description== "max_car_speed":
                self.max_car_speed= int(value[1:])

            elif len(description) > 9 and description[0:10]== "bonus_time":
                self.bonus_time_arr.append( float(value[1:]) )


            elif len(description) > 6 and description== "floor 1":
                i+= 1
                while i< len(line_arr):
                    colin= line_arr[i].find(":")
                    if colin != -1:    # moved on to another description, value pair
                        i-= 1
                        break
                    tile_row= map(lambda(x):int(x), string.split(line_arr[i], " "))
                    self.tile_floor1_arr2d.append( tile_row )
                    i+= 1

            elif len(description) > 6 and description== "floor 2":
                i+= 1
                while i< len(line_arr):
                    colin= line_arr[i].find(":")
                    if colin != -1:    # moved on to another description, value pair
                        i-= 1
                        break
                    tile_row= map( lambda(x):int(x), string.split(line_arr[i], " ") )
                    self.tile_floor2_arr2d.append( tile_row )
                    i+= 1

            elif len(description) > 6 and description== "floor s":
                i+= 1
                while i< len(line_arr):
                    colin= line_arr[i].find(":")
                    if colin != -1:    # moved on to another description, value pair
                        i-= 1
                        break
                    tile_row= map( lambda(x):int(x), string.split(line_arr[i], " ") )
                    self.tile_floor_s_arr2d.append( tile_row )
                    i+= 1

            else:
                _unknown_commands_arr.append(line_arr[i])
                #print "Unknow level commands: ", _unknown_commands_arr

              
        for row in range( 0, len(self.tile_floor_s_arr2d) ):
            for col in range( 0, len(self.tile_floor_s_arr2d[row]) ):
                tile= self.tile_floor_s_arr2d[row][col]
                if tile == helpers.SP_STAR:
                    self.player_start_pos= [self.tile_dimen*col+ self.tile_dimen//2,\
                                       self.tile_dimen*row+self.tile_dimen//2]
                if tile == helpers.SP_CHECK_STAR:
                    self.player_checkpoint_pos= [self.tile_dimen*col+ self.tile_dimen//2,\
                                       self.tile_dimen*row+self.tile_dimen//2]
                elif tile == helpers.SP_MAILBOX1: self.mailbox_pos_arr[0]= [col,row]
                elif tile == helpers.SP_MAILBOX2: self.mailbox_pos_arr[1]= [col,row]
                elif tile == helpers.SP_MAILBOX3: self.mailbox_pos_arr[2]= [col,row]
                elif tile == helpers.SP_MAILBOX4: self.mailbox_pos_arr[3]= [col,row]
                elif tile == helpers.SP_MAILBOX5: self.mailbox_pos_arr[4]= [col,row]

                elif tile == helpers.SP_PARCEL1: self.parcel_pos_arr[0]= [col,row]
                elif tile == helpers.SP_PARCEL2: self.parcel_pos_arr[1]= [col,row]
                elif tile == helpers.SP_PARCEL3: self.parcel_pos_arr[2]= [col,row]
                elif tile == helpers.SP_PARCEL4: self.parcel_pos_arr[3]= [col,row]
                elif tile == helpers.SP_PARCEL5: self.parcel_pos_arr[4]= [col,row]

                elif tile == helpers.SP_CAR_GEN_N: self.cargen_arr.append( Cargen(self, [col, row], helpers.NORTH ) )
                elif tile == helpers.SP_CAR_GEN_E: self.cargen_arr.append( Cargen(self, [col, row], helpers.EAST ) )
                elif tile == helpers.SP_CAR_GEN_S: self.cargen_arr.append( Cargen(self, [col, row], helpers.SOUTH ) )
                elif tile == helpers.SP_CAR_GEN_W: self.cargen_arr.append( Cargen(self, [col, row], helpers.WEST ) )

        # check to make sure all times are posted for bonus money
        while len(self.mailbox_pos_arr) > len(self.bonus_time_arr):
            self.bonus_time_arr.append(100)

        # Add mailbox and parcel to filelist and display the first pair
        self.tile_filename_arr.append(os.path.join(self.tiledir, "mailbox.PNG"))
        self.tile_filename_arr.append(os.path.join(self.tiledir, "parcel.PNG"))
        self.display_mail_objs()

        for parcel in self.parcel_pos_arr:
            if parcel!= [-1, -1]: self.parcel_count+= 1


        ### load all tile images from filenames
        for i in range( 0, len(self.tile_filename_arr) ):
            self.tile_imgnrect_arr.append( self.load_image(self.tile_filename_arr[i], 0xFF0080) )

        fin.close()

        self.width=  len(self.tile_floor1_arr2d[0])* self.tile_dimen
        self.height= len(self.tile_floor1_arr2d)*    self.tile_dimen


        ### load all car images from filenames
        for i in range( len(helpers.car_filenames) ):
            self.car_imgnrect_arr.append( helpers.loadCarImages(helpers.car_filenames[i]) )

 
        ### set up car collision floor
        for i in range( 1+ (len(self.tile_floor1_arr2d)/ self.car_floor_fraction) ):
            self.car_floor_arr2d.append( [[0]]* (1+ (len(self.tile_floor1_arr2d[0])/ self.car_floor_fraction)) )

        for i in range(0, initial_updates):
            self.update()
            
        return




    def load_image(self, fullname, colorkey= None):
        try:
            image= pygame.image.load(fullname)
        except pygame.error, message:
            print 'Level.py: Cannot load image:', fullname
            raise SystemExit, message
        image= image.convert()					   # convert to current color depth
        if colorkey is not None:
            if colorkey is -1:
                colorkey= image.get_at( (0, 0) )
            image.set_colorkey(colorkey, RLEACCEL) # RLEACCEL: slow mod, quick blit
        return image, image.get_rect()




    def update(self):
            self.frame_number+= 1
            self.parcel_delivery_time+= 1
            self.parcel_pick_up_time+= 1

            # update car generation
            for i in range( 0, len(self.cargen_arr) ):
                self.cargen_arr[i].update()
                if self.cargen_arr[i].isTimeForCar():
                    self.cargen_arr[i].resetCounter()
                    pos_vec= [self.cargen_arr[i].get_pos()[0]* self.tile_dimen+ self.tile_dimen/2, \
                              self.cargen_arr[i].get_pos()[1]* self.tile_dimen+ self.tile_dimen/2]
                    direction= self.cargen_arr[i].get_direction()

                    self.car_arr.append( Car(pos_vec, direction, random.randint( self.min_car_speed, self.max_car_speed )) )


            # update cars
            i= 0
            number_of_cars= len(self.car_arr)
            while i< number_of_cars:
                self.car_arr[i].update()
                if self.car_arr[i].sentenced_to_death:
                    number_of_cars-= 1     
                    self.car_arr= self.car_arr[0: i]+ self.car_arr[i+ 1:]
                elif self.car_arr[i].isOB( self.width, self.height ): # remove (delete) O.B. cars
                    number_of_cars-= 1     
                    self.car_arr= self.car_arr[0: i]+ self.car_arr[i+ 1:]
                else:
                    i+= 1


            # car-to-car collision detetection
            for i in range( len(self.car_arr) ):
                x, y= self.car_arr[i].get_x(), self.car_arr[i].get_y()
                col, row= self.get_tile_pos_at_point(x, y)
                row= row/ self.car_floor_fraction
                col= col/ self.car_floor_fraction
                #print "row, col= ", row, col

                status_num= self.car_floor_arr2d[row][col][0]

                if status_num != self.frame_number:
                    # the car collision tile is not up to date: start fresh with the new frame and car
                    self.car_floor_arr2d[row][col]= [self.frame_number, self.car_arr[i]]
                else:

                    for car in self.car_floor_arr2d[row][col][1:]:

                        if self.car_arr[i].grapprect.colliderect(car.grapprect): # car collision!

                            if self.car_arr[i].image_index == car.image_index:

                                speed_difference= abs(self.car_arr[i].speed- car.speed)
                                half_difference= speed_difference/ 2+ 1
                                if self.car_arr[i].speed >= car.speed:
                                    ### current car needs to slow down
                                    lower_speed= max(self.car_arr[i].speed- half_difference, self.min_car_speed)
                                    self.car_arr[i].set_speed( random.randint(self.min_car_speed, lower_speed) )
                                    self.car_arr[i].go_back()
                                    self.car_arr[i].update()

                                    ### other car needs to speed up
                                    greater_speed= min(self.car_arr[i].speed+ half_difference, self.max_car_speed)
                                    car.set_speed( random.randint(greater_speed, self.max_car_speed) )
                                    car.go_back()
                                    car.update()

                                else:
                                    ### current car needs to speed up
                                    greater_speed= min(self.car_arr[i].speed+ half_difference, self.max_car_speed)
                                    self.car_arr[i].set_speed( random.randint(greater_speed, self.max_car_speed) )
                                    self.car_arr[i].go_back()
                                    self.car_arr[i].update()
                                    
                                    ### other car needs to slow down
                                    lower_speed= max(self.car_arr[i].speed- half_difference, self.min_car_speed)
                                    car.set_speed( random.randint(self.min_car_speed, lower_speed) )
                                    car.go_back()
                                    car.update()

                            elif self.car_arr[i].grapprect.colliderect(car.crashrect):
                                #cars collided from different directions: crash and burn!
                                self.car_arr[i].sentence_to_death()
                                car.sentence_to_death()
                                self.explosions_arr2d.append([self.car_arr[i].get_x(), \
                                           self.car_arr[i].get_y(), self.explosion_size, \
                                           helpers.create_explosion(35, min_len= 17, \
                                           low_scale= .4, sections= 10), (208, 189, 51)])
                                self.explosions_arr2d.append([car.get_x(), car.get_y(),  self.explosion_size, 
                                           helpers.create_explosion(35, min_len= 17, low_scale= .4, sections= 12),
                                           (255, 255, 128)])
                                
                    # add the car to the collision detection tile
                    self.car_floor_arr2d[row][col].append( self.car_arr[i] )


            # update explosions
            i= len(self.explosions_arr2d)- 1
            while i >= 0:
                self.explosions_arr2d[i][2]-= 1
                if self.explosions_arr2d[i][2] < 0:
                    self.explosions_arr2d= self.explosions_arr2d[0: i]+ self.explosions_arr2d[i+ 1:]
                i-= 1



    def draw_floor(self, floor_arr2d, tile_imgnrect_arr, background):

        start_x_tile= self.screen_pos[0]// self.tile_dimen
        if start_x_tile < 0: start_x_tile= 0
        end_x_tile= start_x_tile+ self.screen_size[0]// self.tile_dimen+ 1
        if end_x_tile >= len(floor_arr2d[0]): end_x_tile= len(floor_arr2d[0])
        start_x_tile= int(start_x_tile)
        end_x_tile= int(end_x_tile)

        start_y_tile= self.screen_pos[1]// self.tile_dimen
        if start_y_tile < 0: start_y_tile= 0
        end_y_tile= start_y_tile+ self.screen_size[1]// self.tile_dimen+ 2
        if end_y_tile >= len(floor_arr2d): end_y_tile= len(floor_arr2d)
        start_y_tile= int(start_y_tile)
        end_y_tile= int(end_y_tile)

        for y in range(start_y_tile, end_y_tile):
            for x in range(start_x_tile, end_x_tile):
                if floor_arr2d[y][x] != 0:   # do not draw blank tiles
                    tile_image= tile_imgnrect_arr[ floor_arr2d[y][x] ][0]
                    tile_rect=  tile_imgnrect_arr[ floor_arr2d[y][x] ][1]
          
                    tile_rect.topleft = [x*self.tile_dimen-self.screen_pos[0], \
                                         y*self.tile_dimen-self.screen_pos[1]]
                    background.blit(tile_image, tile_rect)





    def draw_mail_objs(self, background):
        # don't display next parcel until one being held is delivered
        if self.cur_parcel == self.cur_mailbox:
            parcel= self.tile_imgnrect_arr[len(self.tile_imgnrect_arr)- 1]
            x= self.parcel_pos_arr[self.cur_parcel][0]
            y= self.parcel_pos_arr[self.cur_parcel][1]
            parcel[1].topleft = [x*self.tile_dimen- self.screen_pos[0], \
                                 y*self.tile_dimen-self.screen_pos[1]]
            background.blit(parcel[0], parcel[1])

        mailbox= self.tile_imgnrect_arr[len(self.tile_imgnrect_arr)- 2]
        x= self.mailbox_pos_arr[self.cur_mailbox][0]
        y= self.mailbox_pos_arr[self.cur_mailbox][1]
        mailbox[1].topleft = [x*self.tile_dimen- self.screen_pos[0], \
                           y*self.tile_dimen-self.screen_pos[1]]
        background.blit(mailbox[0], mailbox[1])





    def draw_cars(self, background):
        if self.car_arr != [] and self.car_arr != None:
            for i in range( 0, len(self.car_arr) ):
                car=  self.car_arr[i]
                rect= car.get_pos()

                car.draw(background, self.car_imgnrect_arr[car.file_number], \
                        rect[0]- self.screen_pos[0], rect[1]-self.screen_pos[1])
        
        # draw explosions
        for explosion in self.explosions_arr2d:
            boom= explosion[3][:]
            for i in range( len(boom) ):
                boom[i]= helpers.sub_points( boom[i], [self.screen_pos[0], self.screen_pos[1]])
                boom[i]= helpers.add_points( boom[i], [explosion[0],explosion[1]])
            for i in range( len(explosion[3]) ):
                explosion[3][i]= helpers.scale( explosion[3][i], 1.0+ self.explosion_size_inc )

            pygame.draw.polygon( background, explosion[4], boom)
            pygame.draw.polygon( background, (128, 128, 64), boom, 2)





    def invalidate_bonus_potential(self):
        self.can_get_bonus_money= False

    def can_get_bonus(self):
        return self.can_get_bonus_money



    def get_special_at_pos(self, pos):
        _x= pos[0]
        _y= pos[1]
        tile_pos= self.get_tile_pos_at_point(_x, _y)

        return self.tile_floor_s_arr2d[ tile_pos[1] ][ tile_pos[0] ]



    def get_tile_pos_at_point(self, _x, _y):
        col_len= len(self.tile_floor_s_arr2d)
        row_len= len(self.tile_floor_s_arr2d[0])

        row= int(_x// self.tile_dimen)
        col= int(_y// self.tile_dimen)

        if row >= row_len: row= row_len- 1
        if col >= col_len: col= col_len- 1
        if row < 0: row= 0
        if col < 0: col= 0

        return row, col




    def display_mail_objs(self):
        pass            # no longer necessary

    def pick_up_parcel(self):
        self.parcel_delivery_time= 0
        cur_parcel_pos= self.parcel_pos_arr[self.cur_parcel]
        x= cur_parcel_pos[0]
        y= cur_parcel_pos[1]
        self.cur_parcel+= 1
        self.display_mail_objs()


    def drop_off_parcel(self):
        self.can_get_bonus_money= True
        self.parcel_pick_up_time= 0
        cur_mailbox_pos= self.mailbox_pos_arr[self.cur_mailbox]
        x= cur_mailbox_pos[0]
        y= cur_mailbox_pos[1]
        self.cur_mailbox+= 1
        self.display_mail_objs()



    def get_parcel_loc(self):
        return [  self.parcel_pos_arr[self.cur_parcel][0]* self.tile_dimen, \
                  self.parcel_pos_arr[self.cur_parcel][1]* self.tile_dimen  ]

    def get_mailbox_loc(self):
        return [  self.mailbox_pos_arr[self.cur_mailbox][0]* self.tile_dimen, \
                  self.mailbox_pos_arr[self.cur_mailbox][1]* self.tile_dimen  ]


    def finished_delivering(self):
        if self.cur_mailbox >= self.parcel_count:
            return True
        else:
            return False


    def is_at_parcel(self, pos):
        cur_parcel_pos= self.parcel_pos_arr[self.cur_parcel]
        player_possie= self.get_tile_pos_at_point(pos[0],pos[1] )
        if player_possie[0] == cur_parcel_pos[0] and \
           player_possie[1]== cur_parcel_pos[1] and\
           self.cur_parcel == self.cur_mailbox: # one being held is still needs delivery
            return True
        else:
            return False


    def is_at_mailbox(self, pos):
        cur_mailbox_pos= self.mailbox_pos_arr[self.cur_mailbox]
        player_possie= self.get_tile_pos_at_point(pos[0],pos[1] )
        if player_possie[0] == cur_mailbox_pos[0] and \
           player_possie[1]== cur_mailbox_pos[1] and \
           self.cur_parcel != self.cur_mailbox:
            return True
        else:
            return False


    def set_screen_pos(self, pos):
        self.screen_pos= pos

    def set_screen_center_pos(self, pos):
        self.screen_pos[0]= pos[0]- self.screen_size[0]// 2
        self.screen_pos[1]= pos[1]- self.screen_size[1]// 2


    def get_next_level(self):
        return self.next_level



