# Generic game object class For Kourier
# Ben Delacour Spring 2006

from __future__ import division

import glob
import helpers
import math
import os, os.path
import pygame
import string
import sys
from GameObject import GameObject
from pygame.locals import *




class Player(GameObject):
    """A generic game object class for Kourier"""

    def __init__( self, pos_vec= [0, 0], mov_vec= [0, 0] ):
        GameObject.__init__(self, pos_vec, mov_vec)

        self.img_arr= self.loadBoarderImages()
        self.grapple_img_arr= self.loadGrappleImages()
        self.image_index= 0
        self.has_parcel= False
        self.money= 0
        self.money_str= "$0.00"
        self.start_lives= 3
        self.lives= self.start_lives

        self.reg_speed= 10.0        # regular speed when not grappled
        self.high_speed_dec= -.12   # high speed decrease
        self.speed_dec= -.3         # speed decrease when moving at low speeds
        self.brake_speed_dec= -1
        self.brake_high_speed_dec= -.64
        self.brakes= False
        self.on_OK_terrain= False
        self.OK_speed_dec= -.179

        self.damage= 0.0            # Dead when >= 1.0
        self.crashrect= Rect(0, 0, 10, 10)
        self.invincible_duration= 6
        self.invincible_time= 0
        self.invincible_distance= 10
        self.location_hurt= [-10, -10]

        self.shooting_speed=    42.0
        self.shooting_direction= 0
        self.retracting_speed= -26.0
        self.shooting=    False
        self.grappled=    False
        self.grapple_vec= [0, 0]
        self.grappled_car_speed= 0
        self.grappled_rotation_angle= math.pi/30
        self.grappled_reel_speed= 4
        self.grapple_max_len= 164
        self.grapple_min_len= 40
        self.grapple_buffer= self.grappled_reel_speed+ .5

        self.left=  False
        self.right= False
        self.up=    False
        self.down=  False
                                         ### extended directions directly represent state of keys
        self.left_ex=  False
        self.right_ex= False
        self.up_ex=    False
        self.down_ex=  False




    def Reset(self, pos_vec):
        self.damage= 0.0
        self.pos_vec= pos_vec
        self.mov_vec= [0, 0]
        self.shooting=    False
        self.grappled=    False
        self.grapple_vec= [0, 0]
        self.on_OK_terrain= False


    def game_over_reset(self, pos_vec):
        self.money= 0
        self.money_str= "$0.00"
        self.Reset(pos_vec)
        self.brakes= False
        self.lives= self.start_lives
        self.left= self.right= self.up= self.down= False
        self.has_parcel= False



    def get_index(self):
        return self.image_index


    def get_shooting_index(self):
        return self.shooting_direction



    def get_grapple_world_pt(self):
        return [self.grapple_vec[0]+ self.pos_vec[0], \
                self.grapple_vec[1]+ self.pos_vec[1]]


    def get_half_grapple_world_pt(self):
        return [(self.grapple_vec[0]/ 2)+ self.pos_vec[0], \
                (self.grapple_vec[1]/ 2)+ self.pos_vec[1]]


    def set_speed(self, speed):
        deg45=    0.7071067812
        factor45= speed* deg45
        if self.up:    self.mov_vec= [0, -speed];
        if self.down:  self.mov_vec= [0, speed]; 
        if self.left:  self.mov_vec= [-speed, 0];
        if self.right: self.mov_vec= [speed, 0];
        if self.right and self.up:   self.mov_vec= [factor45, -factor45]; 
        if self.up and self.left:    self.mov_vec= [-factor45, -factor45];
        if self.left and self.down:  self.mov_vec= [-factor45, factor45];
        if self.down and self.right: self.mov_vec= [factor45, factor45];



    def make_invincible(self):
        self.invincible_time= self.invincible_duration


    def is_invincible(self):
        if self.invincible_time > 0:
            return True
        return False




    def shared_update(self):
        if self.invincible_time > 0:
            self.invincible_time-= 1
        self.crashrect.center=   self.pos_vec



    def update(self):
        self.shared_update()

        terrain_decrease= 0
        if self.on_OK_terrain:
            terrain_decrease= self.OK_speed_dec

        if not self.grappled:
            cur_speed= math.hypot(self.mov_vec[0], self.mov_vec[1])
            speed= self.reg_speed+ terrain_decrease* 15
            if self.brakes or cur_speed > self.reg_speed+ .03: # braking or fast drifting
                speed= cur_speed


            deg45=    0.7071067812
            factor45= speed* deg45
            active= False
            if self.up:    self.mov_vec= [0, -speed]; active= True
            if self.down:  self.mov_vec= [0, speed]; active= True
            if self.left:  self.mov_vec= [-speed, 0]; active= True
            if self.right: self.mov_vec= [speed, 0]; active= True
            if self.right and self.up:   self.mov_vec= [factor45, -factor45]; active= True
            if self.up and self.left:    self.mov_vec= [-factor45, -factor45]; active= True
            if self.left and self.down:  self.mov_vec= [-factor45, factor45]; active= True
            if self.down and self.right: self.mov_vec= [factor45, factor45]; active= True


            if self.brakes:
                unit_vec=   helpers.make_unit(self.mov_vec)
                speed_dec=  self.brake_speed_dec+ terrain_decrease
                if cur_speed > self.reg_speed:
                    speed_dec= self.brake_high_speed_dec+ terrain_decrease
                scaled_vec= helpers.scale(unit_vec, speed_dec)
                if math.hypot(scaled_vec[0], scaled_vec[1]) < cur_speed:
                    self.add_point_to_mov(scaled_vec)
                    speed= cur_speed
                else:
                    self.mov_vec= [0, 0]
                    speed= 0

            else:
                ### decrease speed
                unit_vec=  helpers.make_unit(self.mov_vec)
                speed_dec= self.speed_dec+ terrain_decrease
                if cur_speed > self.reg_speed:
                    speed_dec= self.high_speed_dec+ terrain_decrease
                scaled_vec= helpers.scale(unit_vec, speed_dec)
                if math.hypot(scaled_vec[0], scaled_vec[1]) > cur_speed:
                    scaled_vec= [0, 0]  # stop annoying slight drift
                    if not active: self.mov_vec= [0, 0]
                self.add_point_to_mov(scaled_vec)


            dirr= self.get_direction_from_flags()
            if dirr != -1: self.image_index= dirr
            
            if self.grapple_vec != [0, 0]:  ### grapple is out, animate it
                if not self.grappled:
                    if self.shooting:
                        increase= helpers.make_scaled_unit(self.grapple_vec, self.shooting_speed)
                        grapple_len= math.hypot(self.grapple_vec[0], self.grapple_vec[1])
                        if grapple_len < self.grapple_max_len:
                            self.grapple_vec= helpers.add_points(self.grapple_vec, increase)
                        else:
                            self.grapple_vec= helpers.scale(self.grapple_vec, \
                                              self.grapple_max_len/ grapple_len)
                            self.shooting= False
                            
                    else: 
                        decrease= helpers.make_scaled_unit(self.grapple_vec, self.retracting_speed)
                        if math.hypot(decrease[0], decrease[1]) < math.hypot(self.grapple_vec[0], self.grapple_vec[1]):
                            self.grapple_vec= helpers.add_points(self.grapple_vec, decrease)
                        else:
                            self.grapple_vec= [0, 0]

        else:   # self.grappled
            pass

        GameObject.update(self)




    def get_direction_from_flags(self):
        direction= -1
        if self.up:                  direction= 0
        if self.down:                direction= 4
        if self.left:                direction= 6
        if self.right:               direction= 2
        if self.right and self.up:   direction= 1
        if self.up and self.left:    direction= 7
        if self.left and self.down:  direction= 5
        if self.down and self.right: direction= 3
        return direction



    def is_rel_left(self, base_direction):
        direction= self.get_direction_from_flags()
        if   (base_direction == 0 and self.left_ex) \
          or (base_direction == 1 and self.left_ex and self.up_ex) \
          or (base_direction == 2 and self.up_ex) \
          or (base_direction == 3 and self.left_ex and self.down_ex) \
          or (base_direction == 4 and self.left_ex) \
          or (base_direction == 5 and self.left_ex and self.up_ex) \
          or (base_direction == 6 and self.down_ex) \
          or (base_direction == 7 and self.left_ex and self.down_ex): return True
        else: return False


    def is_rel_right(self, base_direction):
        direction= self.get_direction_from_flags()
        if   (base_direction == 0 and self.right_ex) \
          or (base_direction == 1 and self.right_ex and self.down_ex) \
          or (base_direction == 2 and self.down_ex) \
          or (base_direction == 3 and self.right_ex and self.up_ex) \
          or (base_direction == 4 and self.right_ex) \
          or (base_direction == 5 and self.right_ex and self.down_ex) \
          or (base_direction == 6 and self.up_ex) \
          or (base_direction == 7 and self.right_ex and self.up_ex): return True
        else: return False


    def is_rel_down(self, base_direction):
        direction= self.get_direction_from_flags()
        if   (base_direction == 0 and self.down_ex) \
          or (base_direction == 1 and self.down_ex and self.left_ex) \
          or (base_direction == 2 and self.left_ex) \
          or (base_direction == 3 and self.left_ex and self.up_ex) \
          or (base_direction == 4 and self.up_ex) \
          or (base_direction == 5 and self.up_ex and self.right_ex) \
          or (base_direction == 6 and self.right_ex) \
          or (base_direction == 7 and self.right_ex and self.down_ex): return True
        else: return False


    def is_rel_up(self, base_direction):
        direction= self.get_direction_from_flags()
        if   (base_direction == 0 and self.up_ex) \
          or (base_direction == 1 and self.up_ex and self.right_ex) \
          or (base_direction == 2 and self.right_ex) \
          or (base_direction == 3 and self.right_ex and self.down_ex) \
          or (base_direction == 4 and self.down_ex) \
          or (base_direction == 5 and self.down_ex and self.left_ex) \
          or (base_direction == 6 and self.left_ex) \
          or (base_direction == 7 and self.left_ex and self.up_ex): return True
        else: return False



    def get_rel_left(self, base_direction):
        if   base_direction == 0: return 7
        elif base_direction == 1: return 0
        elif base_direction == 2: return 1
        elif base_direction == 3: return 2
        elif base_direction == 4: return 5
        elif base_direction == 5: return 6
        elif base_direction == 6: return 5
        elif base_direction == 7: return 6

    def get_rel_right(self, base_direction):
        if   base_direction == 0: return 1
        elif base_direction == 1: return 2
        elif base_direction == 2: return 3
        elif base_direction == 3: return 2  
        elif base_direction == 4: return 3
        elif base_direction == 5: return 4
        elif base_direction == 6: return 7
        elif base_direction == 7: return 0




    def rotation_factor(self, base_direction):
        # used to reverse rotation when heading downwards for a more intuitive feel
        if base_direction == 3 or \
           base_direction == 4 or \
           base_direction == 5: return -1
        else: return 1





    def update_grappled(self, car, level):
        self.shared_update()

        self.grappled_car_speed= car.get_speed()
        prev_pos= self.pos_vec

        is_rel_left=  self.is_rel_left(car.image_index)
        is_rel_right= self.is_rel_right(car.image_index)


        ### rotate left and right manually (both relative to car)
        if is_rel_left:    self.image_index= self.get_rel_left(car.image_index)
        elif is_rel_right: self.image_index= self.get_rel_right(car.image_index)
        else:              self.image_index= car.image_index

        rotation_angle=  self.grappled_rotation_angle* self.rotation_factor(car.image_index)
        if is_rel_left:  self.grapple_vec= helpers.rotate_point(self.grapple_vec, rotation_angle)
        if is_rel_right: self.grapple_vec= helpers.rotate_point(self.grapple_vec, -rotation_angle)


        # reel in/ out
        if self.is_rel_down(car.image_index):
            if ( math.hypot(self.grapple_vec[0], self.grapple_vec[1]) < self.grapple_max_len ):
                unit_vec=     helpers.make_unit(self.grapple_vec)
                scaled_vec=   helpers.scale(unit_vec, self.grappled_reel_speed)
                self.grapple_vec= helpers.add_points(self.grapple_vec, scaled_vec)

        if self.is_rel_up(car.image_index):
            if ( math.hypot(self.grapple_vec[0], self.grapple_vec[1]) > self.grapple_min_len ):
                unit_vec=     helpers.make_unit(self.grapple_vec)
                scaled_vec=   helpers.scale(unit_vec, self.grappled_reel_speed)
                self.grapple_vec= helpers.sub_points(self.grapple_vec, scaled_vec)

        self.pos_vec= helpers.sub_points(car.pos_vec, self.grapple_vec)


        # X'ed out tile: cannot enter:
        if level.get_special_at_pos( self.get_pos() ) == helpers.SP_X:
            # !!! try only moving y
            if level.get_special_at_pos( self.get_pos_with_old_x() ) != helpers.SP_X:
                self.use_y_movement_only();
            # !!! try only moving x
            elif level.get_special_at_pos( self.get_pos_with_old_y() ) != helpers.SP_X:
                self.use_x_movement_only();
            else:
                self.pos_vec= prev_pos

        self.grapple_vec= helpers.sub_points(car.pos_vec, self.pos_vec)

        # too far away: release
        if math.hypot(self.grapple_vec[0], self.grapple_vec[1])> self.grapple_max_len+ self.grapple_buffer:
            self.release_grapple()


        GameObject.update(self)
                       



    def shoot_grapple(self):
        if self.grapple_vec == [0, 0]:
            self.shooting= True
            self.shooting_direction= self.image_index

            if   self.image_index== 0: self.grapple_vec= [ 0, -1]
            elif self.image_index== 1: self.grapple_vec= [ 1, -1]
            elif self.image_index== 2: self.grapple_vec= [ 1,  0]
            elif self.image_index== 3: self.grapple_vec= [ 1,  1]
            elif self.image_index== 4: self.grapple_vec= [ 0,  1]
            elif self.image_index== 5: self.grapple_vec= [-1,  1]
            elif self.image_index== 6: self.grapple_vec= [-1,  0]
            elif self.image_index== 7: self.grapple_vec= [-1, -1]


    def attach_grapple(self, car, level):
        self.shooting=    False
        self.grappled=    True
        self.mov_vec= [0, 0]
        self.grapple_vec= helpers.sub_points(car.pos_vec, self.pos_vec)

        ### Adjusting from the grapple detection box to the center of the car can make the grapple too long
        ### ... so this fixes that
        grapple_length= math.hypot(self.grapple_vec[0], self.grapple_vec[1])
        if grapple_length > self.grapple_max_len:
            new_grapple_vec= helpers.scale( self.grapple_vec, self.grapple_max_len/ grapple_length)
            new_pos= helpers.sub_points(car.pos_vec, new_grapple_vec)
            if level.get_special_at_pos( new_pos ) != helpers.SP_X:
                self.pos_vec= new_pos
                self.grapple_vec= new_grapple_vec
             
        
            



    def release_grapple(self):
        self.shooting=    False
        self.grappled=    False
        self.grapple_vec= [0, 0]

        speed= self.grappled_car_speed
        deg45=    0.7071067812
        factor45= speed* deg45
        if self.image_index == 0: self.mov_vec= [0, -speed]
        if self.image_index == 4: self.mov_vec= [0, speed]
        if self.image_index == 6: self.mov_vec= [-speed, 0]
        if self.image_index == 2: self.mov_vec= [speed, 0]
        if self.image_index == 1: self.mov_vec= [factor45, -factor45]
        if self.image_index == 7: self.mov_vec= [-factor45, -factor45]
        if self.image_index == 5: self.mov_vec= [-factor45, factor45]
        if self.image_index == 3: self.mov_vec= [factor45, factor45]


    def loadBoarderImages(self):
        # Loads the images of the player
        big_img_name= os.path.join("sprites", "boarder.PNG")
        big_img= helpers.load_image(big_img_name, 0x123456)
        boarder_img_arr= []

        for i in range (0, 8):
            boarder_img_arr.append( pygame.Surface([50, 50]).convert() )
            boarder_img_arr[i].blit( big_img[0], [0, 0], Rect(50*i+ i, 0, 50, 50) )
            boarder_img_arr[i].set_colorkey(0xFF0080, RLEACCEL)

        return boarder_img_arr




    def loadGrappleImages(self):
        # Loads the images of the player
        big_img_name= os.path.join("sprites", "grapple.PNG")
        big_img= helpers.load_image(big_img_name, 0x123456)
        grapple_img_arr= []

        for i in range (0, 8):
            grapple_img_arr.append( pygame.Surface([7, 7]).convert() )
            grapple_img_arr[i].blit( big_img[0], [0, 0], Rect(7*i+ i, 0, 7, 7) )
            grapple_img_arr[i].set_colorkey(0xFF0080, RLEACCEL)

        return grapple_img_arr





    def draw(self, background, x, y, car):

        player_image= self.img_arr[self.image_index]
        player_rect=  self.img_arr[self.image_index].get_rect()

        player_rect.center = (x, y)
        background.blit(player_image, player_rect)

        if self.grapple_vec != [0, 0]:
            grapple_end= helpers.add_points( [x, y], self.grapple_vec )
            pygame.draw.aaline(background, [255, 255, 255], [x, y], grapple_end, True)
            grapple_rect= self.grapple_img_arr[self.image_index].get_rect()
            grapple_rect.center= grapple_end
            grapple_image= self.shooting_direction
            if self.grappled:
                grapple_image= car.image_index
            background.blit(self.grapple_img_arr[grapple_image], grapple_rect)

                        



