# Helper functions For Kourier
# Ben Delacour Spring 2006

import glob
import math
import os, os.path
import pygame
import random
import string
import sys
from pygame.locals import *



def load_image(fullname, colorkey= None):
    try:
        image= pygame.image.load(fullname)
    except pygame.error, message:
        print 'Cannot load image:', fullname
        raise SystemExit, message
    image= image.convert()					   # convert to current color depth
    if colorkey is not None:
        if colorkey is -1:
            colorkey= image.get_at( (0, 0) )
        image.set_colorkey(colorkey, RLEACCEL) # RLEACCEL: slow mod, quick blit
    return image, image.get_rect()


def load_sound(name):
    class NoneSound:
        def play(self): pass
    if not pygame.mixer or not pygame.mixer.get_init():
        return NoneSound()
    fullname = os.path.join('data', name)
    try:
        sound = pygame.mixer.Sound(fullname)
    except pygame.error, message:
        print 'Cannot load sound:', fullname
        #raise SystemExit, message
        return NoneSound()
    return sound



def get_default_font(size):
    # this helper function is included because of potential problems with py2exe
    # mainfont= pygame.font.Font(None, 25)   <-- This may cause problems with py2exe
    return pygame.font.Font( "freesansbold.ttf", int(size* 0.6875) )




def glow_text(background, mainfont, string, x, y, bgcolor, fgcolor, offset):

    text= mainfont.render(string, False, bgcolor)
    textpos= text.get_rect()
    textpos.topright= x, y+ offset          # down
    background.blit(text, textpos)   
    textpos.topright= x, y- offset          # up
    background.blit(text, textpos)   
    textpos.topright= x- offset, y          # left
    background.blit(text, textpos)   
    textpos.topright= x+ offset, y          # right
    background.blit(text, textpos)   
    text= mainfont.render(string, False, fgcolor)
    textpos= text.get_rect()
    textpos.topright= x, y
    background.blit(text, textpos)



def glow_text_ex(background, mainfont, string, x, y, bgcolor, fgcolor, offset, left= True):

    text= mainfont.render(string, True, bgcolor)
    textpos= text.get_rect()
    textpos.topleft= x+ offset, y+ offset  # down, right
    background.blit(text, textpos)
    textpos= text.get_rect()
    textpos.topleft= x- offset, y- offset  # up, left
    background.blit(text, textpos)
    textpos= text.get_rect()
    textpos.topleft= x+ offset, y- offset  # up, right
    background.blit(text, textpos)
    textpos= text.get_rect()
    textpos.topleft= x- offset, y+ offset  # down, left
    background.blit(text, textpos)
    textpos.topleft= x, y+ offset          # down
    background.blit(text, textpos)   
    textpos.topleft= x, y- offset          # up
    background.blit(text, textpos)   
    textpos.topleft= x- offset, y          # left
    background.blit(text, textpos)   
    textpos.topleft= x+ offset, y          # right
    background.blit(text, textpos)   
    text= mainfont.render(string, True, fgcolor)
    textpos= text.get_rect()
    textpos.topleft= x, y
    background.blit(text, textpos)






SP_OH            = 1  
SP_X             = 2
SP_OK            = 3
SP_MAILBOX1      = 4
SP_STAR          = 5
SP_MAILBOX2      = 6
SP_MAILBOX3      = 7
SP_MAILBOX4      = 8
SP_MAILBOX5      = 9
SP_PARCEL1       = 10 
SP_PARCEL2       = 11
SP_PARCEL3       = 12
SP_PARCEL4       = 13
SP_PARCEL5       = 14
SP_DIRECTION_NW  = 15
SP_DIRECTION_NE  = 16
SP_DIRECTION_SE  = 17
SP_DIRECTION_SW  = 18
SP_DIRECTION_N   = 19
SP_DIRECTION_E   = 20
SP_DIRECTION_S   = 21
SP_DIRECTION_W   = 22
SP_CAR_GEN_N     = 23
SP_CAR_GEN_E     = 24
SP_CAR_GEN_S     = 25
SP_CAR_GEN_W     = 26
SP_CHECK_STAR    = 27
SP_CHECK_TRIGGER = 28


NORTH= 0
EAST=  1
SOUTH= 2
WEST=  3

HIGH_SCORE_FILENAME= "high scores.txt"

def get_special_tile_filename_arr():
    return ["zero tile.PNG", "o.PNG", "x.PNG", "ok.PNG", "mailbox 1.PNG", \
         "star.PNG", "mailbox 2.PNG", "mailbox 3.PNG", "mailbox 4.PNG", "mailbox 5.PNG", 
         "parcel 1.PNG", "parcel 2.PNG", "parcel 3.PNG", "parcel 4.PNG",
         "parcel 5.PNG", "nw.PNG", "ne.PNG", "se.PNG", "sw.PNG", "n.PNG", \
         "e.PNG", "s.PNG", "w.PNG", "cg n.PNG", "cg e.PNG", "cg s.PNG", "cg w.PNG", \
         "checkpoint star.PNG", "checkpoint trigger.PNG"]

car_filenames= ["car stripe green.PNG", "car stripe blue.PNG", \
            "car stripe yellow.PNG", "car sport red.PNG", \
            "car sport purple.PNG", "car sport orange.PNG", \
            "car muscle sky blue.PNG", "car muscle yellow.PNG"]


def loadCarImages(filename):
    # Loads the images of the car_img_arr
    big_img_name= os.path.join("sprites", filename)
    big_img= load_image(big_img_name, 0x123456)
    car_img_arr= []

    for i in range (0, 8):
        car_img_arr.append( pygame.Surface([96, 71]).convert() )
        car_img_arr[i].blit( big_img[0], [0, 0], Rect(96*i+ i, 0, 96, 71) )
        car_img_arr[i].set_colorkey(0xFF0080, RLEACCEL)

    return car_img_arr





def rotate_point(point, raddi):
    x= point[0]
    y= point[1]
    x, y= (x * math.cos(raddi)- y * math.sin(raddi), \
           y * math.cos(raddi)+ x * math.sin(raddi))
    return [x, y]

def add_points(point, point2):
    x= point[0]+ point2[0]
    y= point[1]+ point2[1]
    return [x, y]

def sub_points(point, point2):
    x= point[0]- point2[0]
    y= point[1]- point2[1]
    return [x, y]

def distance(point, point2):
    x= point[0]- point2[0]
    y= point[1]- point2[1]
    dist= math.hypot(x, y)
    return dist

def make_unit(point):
    x= point[0]
    y= point[1]
    dist= math.hypot(x, y)
    if dist == 0: return [0, 0]
    x= x/ dist
    y= y/ dist
    return [x, y]

def make_scaled_unit(point, scale):
    x= point[0]
    y= point[1]
    dist= math.hypot(x, y)
    if dist == 0: return [0, 0]
    x= x/ dist
    y= y/ dist
    return [x* scale, y* scale]


def scale(point, scale):
    return [point[0]* scale, point[1]* scale]


def angle(point):
    return -math.atan2(-point[1], point[0])


def intergerize_vec(point):
    return [int(point[0]), int(point[1])]





def create_explosion(max_len= 200, min_len= 100, low_scale= .4, sections= 12):
   
    pie_slice= 2* math.pi/ sections
    poly_orig= [[0, random.randint(min_len, max_len)]]
    for i in range(1, sections):
        poly_orig.append( rotate_point( [0, random.randint(min_len, max_len)], i*pie_slice) )

    for i in range( len(poly_orig) ):
        if i%2 == 1:
            poly_orig[i]= scale(poly_orig[i], low_scale)
          
    return poly_orig






# Adapted from Timothy Downs' text input code
def draw_input_box(background, message1, message2='', text_size=20, box_length= 200, \
                   box_color= (255,0,128), fgcolor1= (128, 128, 128), \
                   fgcolor2= (255,255,255), bgcolor= (0,0,0)):
    "Draws an input box on the background surface"

    fontobject= get_default_font(text_size)
    area= Rect(0, 0, box_length, text_size+ 6)
    area.center= background.get_rect().center
    pygame.draw.rect(background, bgcolor, area, 0)
    pygame.draw.rect(background, box_color, area, 1)
    if len(message1) > 0:
        text= fontobject.render(message1, 1, fgcolor1, bgcolor)
        background.blit(text, area.move(5, 4))
    if len(message2) > 0:
        text= fontobject.render(message2, 1, fgcolor2, bgcolor)

        background.blit(text, area.move(5+ fontobject.size(message1)[0], 4))





