# Kourier Map Editor
# Ben Delacour Spring 2006
# With some code derived from
# Simple Map Editor by Andrew Ulysses Baker


import glob
import helpers
import math
import os, os.path
import pygame
import random
import string
import sys
from pygame.locals import *




def load_image(fullname, colorkey= None):
    try:
        image= pygame.image.load(fullname)
    except pygame.error, message:
        print 'Cannot load image:', fullname
        raise SystemExit, message
    image= image.convert()					   # convert to current color depth
    if colorkey is not None:
        if colorkey is -1:
            colorkey= image.get_at( (0, 0) )
        image.set_colorkey(colorkey, RLEACCEL) # RLEACCEL: slow mod, quick blit
    return image, image.get_rect()




def save_level(_unknown_commands_arr, next_level, _tile_filename_arr, _tile_floor1_arr2d, \
                            _tile_floor2_arr2d, _tile_floor_s_arr2d, filename):

    lev_file= open(filename, "w")

    for line in _unknown_commands_arr:
        lev_file.write( line+ "\n" )

    lev_file.write( "next_level: "+ next_level+ "\n" )

    # write ~"tile x: image.PNG\n" tile filename list
    for i in range(0, len(_tile_filename_arr)):
        outname= os.path.split( _tile_filename_arr[i] )
        lev_file.write( "tile "+ str(i)+ ": " )
        lev_file.write( outname[len(outname)- 1]+ "\n" )

    lev_file.write("floor 1:\n")
    for y in range( 0, len(_tile_floor1_arr2d) ):
        for x in range( 0 , len(_tile_floor1_arr2d[y]) ):
            lev_file.write( str(_tile_floor1_arr2d[y][x]) )
            if x!= len(_tile_floor1_arr2d[y])- 1:
                lev_file.write(" ")
            else:  lev_file.write("\n")

    lev_file.write("floor 2:\n")
    for y in range( 0, len(_tile_floor2_arr2d) ):
        for x in range( 0 , len(_tile_floor2_arr2d[y]) ):
            lev_file.write( str(_tile_floor2_arr2d[y][x]) )
            if x!= len(_tile_floor2_arr2d[y])- 1:
                lev_file.write(" ")
            else:
                lev_file.write("\n")

    lev_file.write("floor s:\n")
    for y in range( 0, len(_tile_floor_s_arr2d) ):
        for x in range( 0 , len(_tile_floor_s_arr2d[y]) ):
            lev_file.write( str(_tile_floor_s_arr2d[y][x]) )
            if x!= len(_tile_floor_s_arr2d[y])- 1:
                lev_file.write(" ")
            else:
                if y!= len(_tile_floor_s_arr2d)- 1:
                    lev_file.write("\n")
            

    lev_file.close()




def load_special_tiles():
    # Loads the special tiles (special tiles are not drawn in game)

    _tile_filename_arr= helpers.get_special_tile_filename_arr()
    _tile_imgnrect_arr= []

    for fname in _tile_filename_arr[0:]:
        fname= os.path.join("special tiles", fname)
        _tile_imgnrect_arr.append( load_image(fname, 0xFF0080) )

    return _tile_filename_arr, _tile_imgnrect_arr




def load_level(_level, tiledir= "tiles"):
    # Loads the specified level

    _unknown_commands_arr= []
    _tile_filename_arr= []
    _tile_imgnrect_arr= []
    _tile_floor1_arr2d= []
    _tile_floor2_arr2d= []
    _tile_floor_s_arr2d= []
    _map_pos_x= 0
    _map_pos_y= 0

    full_name= os.path.join("levels", _level)
    if os.path.exists(full_name)== False:
            print "Error in load_map: map %s does not exist." % full_name
            
    #print "Loading level in load_map: %s" % full_name
    fin= open(full_name, 'r')


    whole= fin.read()
    line_arr= string.split(whole, '\n')
    for i in range(0, len(line_arr)):

        colin= line_arr[i].find(":")
        if colin == -1:
            continue

        [description, value]= string.split(line_arr[i], ":")
        if len(description) > 5 and description[0:4]== "tile":
            _tile_filename_arr.append(os.path.join(tiledir, value[1:len(value)]))
    
        elif len(description) > 9 and description== "next_level":
            next_level= value[1:len(value)]


        elif len(description) > 6 and description== "floor 1":
            #print("Floor 1: ")
            i+= 1
            while i< len(line_arr):
                colin= line_arr[i].find(":")
                if colin != -1:    # moved on to another description, value pair
                    i-= 1
                    break
                tile_row= map(lambda(x):int(x), string.split(line_arr[i], " "))
                _tile_floor1_arr2d.append( tile_row )
                #print(tile_row)
                i+= 1

        elif len(description) > 6 and description== "floor 2":
            #print("Floor 2: ")
            i+= 1
            while i< len(line_arr):
                colin= line_arr[i].find(":")
                if colin != -1:    # moved on to another description, value pair
                    i-= 1
                    break
                tile_row= map( lambda(x):int(x), string.split(line_arr[i], " ") )
                _tile_floor2_arr2d.append( tile_row )
                #print(tile_row)
                i+= 1


        elif len(description) > 6 and description== "floor s":
            #print("Floor S: ")
            i+= 1
            while i< len(line_arr):
                colin= line_arr[i].find(":")
                if colin != -1:    # moved on to another description, value pair
                    i-= 1
                    break
                tile_row= map( lambda(x):int(x), string.split(line_arr[i], " ") )
                _tile_floor_s_arr2d.append( tile_row )
                #print(tile_row)
                i+= 1
        else:
            _unknown_commands_arr.append(line_arr[i])

    for i in range( 0, len(_tile_filename_arr) ):
        _tile_imgnrect_arr.append( load_image(_tile_filename_arr[i], 0xFF0080) )

    #print("")
    #print(_tile_filename_arr)
    #print(_tile_imgnrect_arr)
    #print("")

    fin.close()
    return _unknown_commands_arr, next_level, _tile_filename_arr, _tile_imgnrect_arr, \
                 _tile_floor1_arr2d, _tile_floor2_arr2d, _tile_floor_s_arr2d, \
                                                        _map_pos_x, _map_pos_y

    


def load_all_tiles(directory):
    # Loads all tile graphics in the specified directory
    _all_filename_arr= []
    _all_imgnrect_arr= []

    _all_filename_arr= glob.glob(directory+ "/*.PNG")
    print("\nAll tiles loaded in load_all_tiles: ")
    print(_all_filename_arr)
    print("\n")

    for i in range( 0, len(_all_filename_arr) ):
        _all_imgnrect_arr.append( load_image(_all_filename_arr[i], None) )

    return _all_filename_arr, _all_imgnrect_arr




def load_world_ini(name):
    # Loads world variables from world.ini

    # defaults (if specified)
    _game_screen_x= 640
    _game_screen_y= 480
    _tile_dimen=    64
    _level_start=   "level1.lev"

    if os.path.exists(name) == False:
        print "%s doesn't exist!" % name
    else:
        fin= file(name, 'r')

    whole= fin.read()
    fin.close()
    lines= string.split(whole, '\n')

    for lin in lines:  

        a, b = string.split(lin, ':')
        if a == "game_screen_x":   _game_screen_x =   int(b)
        elif a == "game_screen_y": _game_screen_y =   int(b)
        elif a == "tile_dimen":    _tile_dimen =      int(b)
        elif a == "level_start":   _level_start = b

    return _game_screen_x, _game_screen_y, _tile_dimen, _level_start




def draw_world(tile_imgnrect_arr, tile_floor1_arr2d, tile_floor2_arr2d, \
                               _tile_floor_s_arr2d, screen, background):
    draw_floor(tile_floor1_arr2d, tile_imgnrect_arr, screen, background)
    draw_floor(tile_floor2_arr2d, tile_imgnrect_arr, screen, background)

    if g_shound_render_special:
        draw_floor(_tile_floor_s_arr2d, g_special_tile_imgnrect_arr, screen, background)
      



def draw_floor(floor_arr2d, tile_imgnrect_arr, screen, background):

    start_x_tile= g_screen_pos_x// g_tile_dimen
    if start_x_tile < 0: start_x_tile= 0
    end_x_tile= start_x_tile+ g_game_screen_x// g_tile_dimen+ 1
    if end_x_tile >= len(floor_arr2d[0]): end_x_tile= len(floor_arr2d[0])
    start_x_tile= int(start_x_tile)
    end_x_tile= int(end_x_tile)

    start_y_tile= g_screen_pos_y// g_tile_dimen
    if start_y_tile < 0: start_y_tile= 0
    end_y_tile= start_y_tile+ g_game_screen_y// g_tile_dimen+ 2
    if end_y_tile >= len(floor_arr2d): end_y_tile= len(floor_arr2d)
    start_y_tile= int(start_y_tile)
    end_y_tile= int(end_y_tile)

    for y in range(start_y_tile, end_y_tile):
        for x in range(start_x_tile, end_x_tile):
            #print("tile_imgnrect_arr[ floor_arr2d[x][y] ]")
            #print(tile_imgnrect_arr[ floor_arr2d[x][y] ])
            #print ("drawing tile #, %s" % floor_arr2d[y][x] )
            if floor_arr2d[y][x] != 0:   # do not draw blank tiles
                tile_image= tile_imgnrect_arr[ floor_arr2d[y][x] ][0]
                tile_rect=  tile_imgnrect_arr[ floor_arr2d[y][x] ][1]
      
                tile_rect.topleft = (x*g_tile_dimen-g_screen_pos_x, y*g_tile_dimen-g_screen_pos_y)
                background.blit(tile_image, tile_rect)




def main():
    pygame.init()
    if not pygame.font: print 'Warning, fonts disabled'
    if not pygame.mixer: print 'Warning, sound disabled'
    
    #global g_screen                   # initialized 
    #global g_background               # initialized
    global g_tile_dimen                # initialized
    global g_cur_bg_tile               # index of the background working tile 
    global g_cur_fg_tile               # index of the foreground working tile 
    global g_screen_pos_x              # screen pos rel to the zero of the map
    global g_screen_pos_y              # screen pos rel to the zero of the map
    global g_game_screen_x             # initialized 
    global g_game_screen_y             # initialized
    global g_shound_render_special     # editor specific
    global g_special_tile_filename_arr # editor specific
    global g_special_tile_imgnrect_arr # editor specific

    g_screen_pos_x=  320    # Arbitrary assignment to get away from the edge
    g_screen_pos_y=  320    # Arbitrary assignment to get away from the edge
    g_cur_bg_tile=   0
    g_cur_fg_tile=   0
    g_cur_sp_tile=   0

    g_game_screen_x, \
    g_game_screen_y, \
    g_tile_dimen, \
    cur_level= load_world_ini('world.ini')

    screen_size=         [g_game_screen_x+ g_tile_dimen, g_game_screen_y]
    screen=              pygame.display.set_mode(screen_size)
    background=          pygame.Surface(screen_size).convert()
    pygame.display.set_caption('Kourier Map Editor')

    tile_allfilename_arr, \
    tile_allimgnrect_arr= load_all_tiles(directory= 'tiles')

    unknown_commands_arr, \
    next_level, \
    tile_filename_arr, \
    tile_imgnrect_arr, \
    tile_floor1_arr2d, \
    tile_floor2_arr2d, \
    tile_floor_s_arr2d, \
    map_pos_x, \
    map_pos_y= load_level(cur_level)

    g_special_tile_filename_arr, g_special_tile_imgnrect_arr= load_special_tiles()

    center_x= g_game_screen_x// 2
    center_y= g_game_screen_y// 2
    mouse_x= 1
    mouse_y= 1

    running= True
    moving= True
    eyedropping= False
    g_shound_render_special= True
    while running == True:
        pygame.time.delay(40)

        if pygame.mouse.get_focused():
            mouse_x= pygame.mouse.get_pos()[0]
            mouse_y= pygame.mouse.get_pos()[1]

            # scrolling movement
            if moving:
                dist= math.hypot((center_x- mouse_x),(center_y- mouse_y))
                circ_dist= dist- 2*g_tile_dimen
                if circ_dist <= 0: dist= 0
                rise= center_y- mouse_y
                run=  center_x- mouse_x

                if mouse_x < g_game_screen_x and dist > g_tile_dimen*2:
                    g_screen_pos_x-= ((run/dist)*.6)*circ_dist
                    g_screen_pos_y-= ((rise/dist)*.9)*circ_dist


            if pygame.mouse.get_pressed()[0] == True:   # Change bg tile
                row, col= get_tile_pos_at_point(mouse_x, mouse_y, \
                          tile_floor1_arr2d, g_screen_pos_x, g_screen_pos_y)
                if eyedropping == True:
                    g_cur_bg_tile= tile_floor1_arr2d[col][row]
                    eyedropping= False
                tile_floor1_arr2d[col][row]= g_cur_bg_tile


            if pygame.mouse.get_pressed()[1] == True:   # Change special tile
                row, col= get_tile_pos_at_point(mouse_x, mouse_y, \
                          tile_floor_s_arr2d, g_screen_pos_x, g_screen_pos_y)
                if eyedropping == True:
                    g_cur_sp_tile= tile_floor_s_arr2d[col][row]
                    eyedropping= False
                tile_floor_s_arr2d[col][row]= g_cur_sp_tile


            elif pygame.mouse.get_pressed()[2] == True:   # Change bg tile
                row, col= get_tile_pos_at_point(mouse_x, mouse_y, \
                          tile_floor1_arr2d, g_screen_pos_x, g_screen_pos_y)
                if eyedropping == True:
                    g_cur_fg_tile= tile_floor2_arr2d[col][row]
                    eyedropping= False
                tile_floor2_arr2d[col][row]= g_cur_fg_tile

        for event in pygame.event.get():

            if event.type == QUIT:
                running= False
                save_level(unknown_commands_arr, next_level, tile_filename_arr, \
                           tile_floor1_arr2d, tile_floor2_arr2d, \
                           tile_floor_s_arr2d, os.path.join("levels", \
                           cur_level+".bak"))
            elif event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    running= False
                    save_level(unknown_commands_arr, tile_filename_arr, \
                               tile_floor1_arr2d, tile_floor2_arr2d, \
                               tile_floor_s_arr2d, os.path.join("levels", \
                               cur_level+".bak"))

                elif event.key == K_m:
                    if moving: moving= False
                    else:      moving= True


                elif event.key == K_c:

                    col= get_tile_pos_at_point(center_x, center_y, \
                         tile_floor1_arr2d, g_screen_pos_x, g_screen_pos_y)[1]
                    tile_floor1_arr2d=  delete_column(tile_floor1_arr2d, col)
                    tile_floor2_arr2d=  delete_column(tile_floor2_arr2d, col)
                    tile_floor_s_arr2d= delete_column(tile_floor_s_arr2d, col)
                    

                elif event.key == K_r:
                    row= get_tile_pos_at_point(center_x, center_y, \
                         tile_floor1_arr2d, g_screen_pos_x, g_screen_pos_y)[0]
                    tile_floor1_arr2d=  delete_row(tile_floor1_arr2d, row)
                    tile_floor2_arr2d=  delete_row(tile_floor2_arr2d, row)
                    tile_floor_s_arr2d= delete_row(tile_floor_s_arr2d, row)
 
             
                elif event.key == K_v:

                    col= get_tile_pos_at_point(center_x, center_y, \
                         tile_floor1_arr2d, g_screen_pos_x, g_screen_pos_y)[1]
                    tile_floor1_arr2d=  add_column(tile_floor1_arr2d, col)
                    tile_floor2_arr2d=  add_column(tile_floor2_arr2d, col)
                    tile_floor_s_arr2d= add_column(tile_floor_s_arr2d, col)
                    

                elif event.key == K_t:
                    row= get_tile_pos_at_point(center_x, center_y, \
                         tile_floor1_arr2d, g_screen_pos_x, g_screen_pos_y)[0]
                    tile_floor1_arr2d=  add_row(tile_floor1_arr2d, row)
                    tile_floor2_arr2d=  add_row(tile_floor2_arr2d, row)
                    tile_floor_s_arr2d= add_row(tile_floor_s_arr2d, row)


                
                elif event.key == K_a:
                    cur_level= next_level

                    unknown_commands_arr, \
                    next_level, \
                    tile_filename_arr, \
                    tile_imgnrect_arr, \
                    tile_floor1_arr2d, \
                    tile_floor2_arr2d, \
                    tile_floor_s_arr2d, \
                    map_pos_x, \
                    map_pos_y= load_level(next_level)

                elif event.key == K_s:
                    save_level(unknown_commands_arr, next_level, tile_filename_arr, \
                               tile_floor1_arr2d, tile_floor2_arr2d, \
                               tile_floor_s_arr2d, os.path.join("levels", \
                               cur_level))

                elif event.key == K_d:
                    if g_shound_render_special: g_shound_render_special= False
                    else:                       g_shound_render_special= True


                elif event.key == K_g:
                    print "position: ", g_screen_pos_x, ",", g_screen_pos_y


                elif event.key == K_LEFT:
                    g_cur_bg_tile-= 1
                    tile_count= len(tile_imgnrect_arr)
                    if g_cur_bg_tile < 0: g_cur_bg_tile= tile_count- 1

                elif event.key == K_RIGHT:
                    g_cur_bg_tile+= 1
                    tile_count= len(tile_imgnrect_arr)
                    if g_cur_bg_tile >= tile_count: g_cur_bg_tile= 0

                elif event.key == K_UP:
                    g_cur_fg_tile-= 1
                    tile_count= len(tile_imgnrect_arr)
                    if g_cur_fg_tile < 0: g_cur_fg_tile= tile_count- 1

                elif event.key == K_DOWN:
                    g_cur_fg_tile+= 1
                    tile_count= len(tile_imgnrect_arr)
                    if g_cur_fg_tile >= tile_count: g_cur_fg_tile= 0

                elif event.key == K_PAGEUP:
                    g_cur_sp_tile-= 1
                    tile_count= len(g_special_tile_imgnrect_arr)
                    if g_cur_sp_tile < 0: g_cur_sp_tile= tile_count- 1

                elif event.key == K_PAGEDOWN:
                    g_cur_sp_tile+= 1
                    tile_count= len(g_special_tile_imgnrect_arr)
                    if g_cur_sp_tile >= tile_count: g_cur_sp_tile= 0


                elif event.key == K_SPACE:
                    eyedropping= True


        #background.fill(0x000000)
        draw_world(tile_imgnrect_arr, tile_floor1_arr2d, tile_floor2_arr2d, \
                                      tile_floor_s_arr2d, screen, background)
        background.fill(0xCCCCCC, Rect(g_game_screen_x, 0, g_tile_dimen, g_game_screen_y) )

        # Draw current background tile
        current_tile_image= tile_imgnrect_arr[ g_cur_bg_tile ][0]
        current_tile_rect=  tile_imgnrect_arr[ g_cur_bg_tile ][1]
        current_tile_rect.topleft=  (g_game_screen_x, g_tile_dimen)
        background.blit(current_tile_image, current_tile_rect)

        # Draw current foreground tile
        current_tile_image= tile_imgnrect_arr[ g_cur_fg_tile ][0]
        current_tile_rect=  tile_imgnrect_arr[ g_cur_fg_tile ][1]
        current_tile_rect.topleft=  (g_game_screen_x, 2*g_tile_dimen)
        background.blit(current_tile_image, current_tile_rect)

        # Draw current special tile
        current_tile_image= g_special_tile_imgnrect_arr[ g_cur_sp_tile ][0]
        current_tile_rect=  g_special_tile_imgnrect_arr[ g_cur_sp_tile ][1]
        current_tile_rect.topleft=  (g_game_screen_x, 3*g_tile_dimen)
        background.blit(current_tile_image, current_tile_rect)

        if moving:
            pygame.draw.circle(background, 0xFFFFFF, [center_x, center_y], g_tile_dimen*2, 3)

        screen.blit(background, (0,0))

        pygame.display.flip()
    """Exiting Main Loop"""





def delete_column(tile_floor_arr2d, col):
    tile_floor_arr2d= tile_floor_arr2d[:col] + tile_floor_arr2d[col+1:]
    return tile_floor_arr2d


def delete_row(tile_floor_arr2d, row):
    for i in range( len(tile_floor_arr2d) ):
        tile_floor_arr2d[i]= tile_floor_arr2d[i][:row]+ tile_floor_arr2d[i][row+1:]
    return tile_floor_arr2d



def add_column(tile_floor_arr2d, col):
    new_col= []
    for tile in tile_floor_arr2d[col]:
        new_col= new_col+ [tile]
    tile_floor_arr2d= tile_floor_arr2d[:col] + [new_col] + tile_floor_arr2d[col:]
    return tile_floor_arr2d


def add_row(tile_floor_arr2d, row):
    new_row= []
    for i in range( len(tile_floor_arr2d) ):
        tile= tile_floor_arr2d[i][row]
        new_row= new_row+ [tile]    

    for i in range( len(tile_floor_arr2d) ):
        tile_floor_arr2d[i]= tile_floor_arr2d[i][:row]+ [tile_floor_arr2d[i][row]]+ tile_floor_arr2d[i][row:]
    return tile_floor_arr2d





def get_tile_pos_at_point(_x, _y, _tile_floor_arr2d, _screen_pos_x, _screen_pos_y):
    _x+= g_screen_pos_x;     # now in "world space" in relatiion to zero tile
    _y+= g_screen_pos_y;     # now in "world space" in relatiion to zero tile
    col_len= len(_tile_floor_arr2d)
    row_len= len(_tile_floor_arr2d[0])

    row= int(_x// g_tile_dimen)
    col= int(_y// g_tile_dimen)

    if row >= row_len: row= row_len- 1
    if col >= col_len: col= col_len- 1
    if row < 0: row= 0
    if col < 0: col= 0

    return row, col





if __name__ == '__main__': main()






